function [NewCluster, NewDist, NewMustLink, flag_new] = FindNewClustering(Feature, Cluster, CurrMustLink, CurrCannotLink, AllMustLink, para)
% Find a new clustering such that the intra-class distance is decreased and
% the number of currently satisfied must-links are increased.
% 
% Input:
%   - Feature           : each row is a feature vector
%   - Cluster           : each row is a cluster label
%   - CurrMustLink      : currently satisfied must links
%   - CurrCannotLink    : currently satisfied cannot links
%   - AllMustLink       : all must-links that the ideal clustering would satisfy
%   - para
%       - DistType      : 'Euclidean' (default) or 'Mahalanobis' distance
% Output:
%   - NewCluster        : each row is a new cluster label
%   - NewDist           : the new intra-class distance
%   - NewMustLink       : newly satisfied must links
%   - flag_new      : flag to indicate if the clustering returned is new or not
% 
% Author: Zhiyao Duan
% Created: 6/20/2012
% Last modified: 6/22/2012

% Calculate the current intra-class distance
CurrDist = CalculateIntraClassDist(Feature, Cluster, para.DistType);

% Loope to find an swap set
instNum = size(Feature, 1);
IsInstUsed = zeros(instNum, 1);
initInst = ceil(rand * instNum);                                % randomize the starting notelet
for p = 1:instNum
    % the current instance
    i = initInst-1 + p;
    if i > instNum
        i = mod(i, instNum);
    end    
    
    % check if we need to consider i anymore. If another instance which is
    % in the same must-link group as i has been considered, then i needs
    % not to be considered anymore.
    if IsInstUsed(i) == 1
        continue;
    end
    
    % the cluster of i
    currCluster = Cluster(i);
    
    % find the set of instances that are connected to i through must-links
    currSet = FindInstThroughMustLinks(i, CurrMustLink);
    
    % we have considered i, so we don't need to consider any other
    % instances in currg anymore. We set the flag.
    IsInstUsed(currSet) = 1;
    
    % try all possible target clusters
    uCluster = unique(Cluster)';
    tempDist = ones(1, length(uCluster)) * Inf;
    for q = 1:length(uCluster)
        targCluster = uCluster(q);
        if currCluster == targCluster
            continue;
        end
        
        % find the swap set between the current cluster and the target
        % cluster
        SwapSet = FindSwapSet(currSet, currCluster, targCluster, Cluster, CurrMustLink, CurrCannotLink);
        
        % make the swap
        tempCluster = Cluster;
        idx = Cluster(SwapSet) == currCluster;
        tempCluster(SwapSet(idx)) = targCluster;
        idx = Cluster(SwapSet) == targCluster;
        tempCluster(SwapSet(idx)) = currCluster;
        
        % calculate the new intr-class distance after the swap
        tempDist(q) = CalculateIntraClassDist(Feature, tempCluster, para.DistType);
    end
    
    % find the best swap
    [minDist, idx] = min(tempDist);
    if minDist < CurrDist
        % update the intra-class distance
        NewDist = minDist;
        
        % make the swap
        targCluster = uCluster(idx);
        SwapSet = FindSwapSet(currSet, currCluster, targCluster, Cluster, CurrMustLink, CurrCannotLink);
        NewCluster = Cluster;
        idx = Cluster(SwapSet) == currCluster;
        NewCluster(SwapSet(idx)) = targCluster;
        idx = Cluster(SwapSet) == targCluster;
        NewCluster(SwapSet(idx)) = currCluster;    
                       
        % update the must-links
        NewMustLink = CurrMustLink;
        for j = SwapSet
            for k = AllMustLink{j}
                if NewCluster(j)==NewCluster(k)
                    NewMustLink{j} = unique([NewMustLink{j}, k]);
                    NewMustLink{k} = unique([NewMustLink{k}, j]);
                end
            end
        end
        
        % find a new clustering and return
        flag_new = 1;
        return;
    end
end
% did not find any new clustering and return
flag_new = 0;
NewCluster = Cluster;
NewDist = CurrDist;
NewMustLink = CurrMustLink;