function doMPS_L(wavFile, estF0File, outFile, trackNum, dataType, IfDebug)
% Multi-pitch streaming with only locality info based on random initialization at each frame.
%
% Input:
%   - wavFile       : the audio needs to be processed
%   - estF0File     : the mpe input file, format depends on the filename
%   - outFile       : the mps result file, each row is a frame, each column
%                     is a speaker, values are in MIDI number
%   - trackNum      : the number of speakers
%   - dataType      : 'music' or 'speech'
%   - IfDebug       : 0 - not debug; 1 - debug mode, a number of figures
%                     will be shown
%
% Example usage:
%   doMPS_L('example\01-AchGottundHerr-short.wav', 'example\01-AchGottundHerr-short_est.f0', 'trk.f0', 4, 'music', 0);
% or
%   doMPS_L('example\022_M10_si2182_F02_si803_F10_si2267_cut.wav', 'example\022_M10_si2182_F02_si803_F10_si2267_cut_est.f0', 'trk.f0', 3, 'speech', 0);
%
% Author: Zhiyao Duan
% Created: 9/26/2012
% Last modified: 5/31/2013

fprintf('\nMulti-pitch streaming with locality info based on random initialization at each frame, on file %s...\n', wavFile);

if nargin < 6
    IfDebug = 0;
end
if IfDebug == 1
    close all;
end

%% Set parameters
% The following scripts set all the parameters for the given datatype. One can
% modify these parameters by editing the corresponding script.
switch dataType
    case 'music'
        SetParameters_music;
    case 'speech'
        SetParameters_speech;
    otherwise
        fprintf('Unknown data type!\n');
        return;
end

%% Load multi-pitch estimation results, and calculate STFT, peaks, etc.
fprintf('----------Step 1: load multi-pitch estimation results and perform basic calculations such as STFT, peak detection, etc.----------\n');

% Each row of estF0File corresponds to a frame. Its values are in MIDI number.
EstF0 = dlmread(estF0File);
EstF0 = EstF0';
if size(EstF0,2) > trackNum
    EstF0 = EstF0(1:trackNum, :);
end

% Read the wavfiles
[wavData, fs, ~] = ReadWavFile_Mono(wavFile);
% Downsample if it is speech
if strcmp(dataType, 'speech')
    wavData = resample(wavData,16000,fs);
    fs = 16000;
end
% wavData = wavData(1:fs*5);                                          % TEMP: chunk first 5 seconds for debugging

% Detect silent frames
para = {};
para.frameLen = frameLen;
para.hop = hop;
para.win = win;
para.RMS_th = RMS_th;
FrameIndex = DetectSilentFrames(wavData, para);
if IfDebug == 1
    % plot FrameIndex
    figure; plot(FrameIndex); ylim([0 2]);
end

% STFT
fftLen = 2^nextpow2(frameLen*zpf);                                  % zero padding
para = {};
para.frameLen = frameLen;
para.hop = hop;
para.win = win;
para.fftLen = fftLen;
SpecData = CalculateSTFT(wavData, para);
if IfDebug == 1
    % plot spectrum
    figure;
    imagesc(1:length(FrameIndex), (1:fftLen/2)*fs/fftLen, 20*log10(abs(SpecData(1:fftLen/2, :))));
    set(gca,'YDir','normal');
    title('Spectrogram and ground-truth pitches');
    xlabel('Time (Frame number)');
    ylabel('Frequency (Hz)');
    ylim([0 2000]);
%     % plot the ground-truth pitch
%     % uncomment this block if you have the ground-truth pitches
%     GTF0s = dlmread(strrep(wavFile, '.wav', '.f0'));                % in Hz
%     GTF0s = GTF0s';
%     hold on;
%     plot(1:size(GTF0s,2), GTF0s, '.');
end

% Peak Detection
para = {};
para.peakTh = peakTh;
para.peakTh_rel = peakTh_rel;
para.peakTh_freq = min(fftLen/2, round(peakTh_freq*fftLen/fs));     % frequency threshold (point)
para.movL = round(movL*fftLen/fs);                                  % moving average length (point)
para.localRange = localRange;
[PeakData, PeakAmpData, PeakRelAmpData, PeakNum] = CalculatePeakData(SpecData, FrameIndex, para);
idx = PeakData==0;
PeakData = hz2midi(PeakData*fs/fftLen);                             % change to midi number
PeakData(idx) = 0;
if IfDebug == 1
    % plot peaks
    tempFrame = 293;
    figure; plot((1:fftLen/2)*fs/fftLen, 20*log10(abs(SpecData(1:fftLen/2,tempFrame))));
    hold on; plot(midi2hz(PeakData(1:PeakNum(tempFrame),tempFrame)), PeakAmpData(1:PeakNum(tempFrame),tempFrame), 'ro')
end

%% Multi-pitch Streaming
fprintf('----------Step 2: multi-pitch streaming----------\n');

% Calculate an initial tracking result according to pitch order in each frame
EstF0_InitialTrack = zeros(size(EstF0));
for i = 1:length(FrameIndex)
    EstF0_InitialTrack(:, i) = EstF0(randperm(trackNum), i);         % ramdom initialization at each frame
end
if IfDebug==1
    figure;
    imagesc(1:length(FrameIndex), (1:fftLen/2)*fs/fftLen, 20*log10(abs(SpecData(1:fftLen/2, :))));
    set(gca,'YDir','normal');
    title('Initial tracking');
    xlabel('Time (Frame number)');
    ylabel('Frequency (Hz)');
    ylim([0 2000]);
    hold on;
    plot(1:size(EstF0_InitialTrack,2), midi2hz(EstF0_InitialTrack), '.');
end

% Calculate a feature vector for each F0 in each frame
switch FeatureType
    % calculate a simplified version of harmonic structure feature
    case 'HarmStruct'
        para = {};
        para.maxHarm = maxHarm;
        para.Thresh_PeakF0Belong = Thresh_PeakF0Belong;
        [F0Feature, F0Location, F0Track] = CalculateF0Features_HarmonicStructure(PeakData, PeakAmpData, EstF0_InitialTrack, para);    
    case 'UDC'
        % calculate harmonic discrete cepstrum features
        para = {};
        para.fs = fs;
        para.frameLen = frameLen;
        para.fftLen = fftLen;
        para.maxHarm = maxHarm;
        para.Thresh_PeakF0Belong = Thresh_PeakF0Belong;
        para.cepsOrder = cepsOrder;
        [F0Feature, F0Location, F0Track, F0Harmonics] = CalculateF0Features_UniformDiscreteCepstrum(SpecData, PeakData, PeakAmpData, EstF0_InitialTrack, para);  
        if IfDebug == 1
            tempF0Idx = 3;
            tempFrame = F0Location(tempF0Idx, 1);
            tempPitch = midi2hz(F0Location(tempF0Idx, 2));
            % plot spectrum and peaks
            figure; plot((1:fftLen/2)*fs/fftLen, 20*log10(abs(SpecData(1:fftLen/2,tempFrame))));
            hold on; plot(midi2hz(PeakData(1:PeakNum(tempFrame),tempFrame)), PeakAmpData(1:PeakNum(tempFrame),tempFrame), 'ro')
            % plot F0 and harmonics
            tempHarmonicFreq = tempPitch * (1:maxHarm);
            hold on; plot([tempHarmonicFreq; tempHarmonicFreq], [-40*ones(1,maxHarm); F0Harmonics(tempF0Idx,:)], 'go'); 
            % plot spectral envelope
            hold on; plotSpecEnvFromDisCeps(F0Feature(tempF0Idx,:), fs, fftLen);
            ylim([-40 60]);
            xlim([0 16000]);
        end
    case 'MFCC'
        % calculate MFCC feature
        para = {};
        para.fs = fs;
        para.frameLen = frameLen;
        para.fftLen = fftLen;
        para.maxHarm = maxHarm;
        para.cepsOrder = cepsOrder;
        para.trackNum = trackNum;
        para.harmOvlpTh = harmOvlpTh;
        [F0Feature, F0Location, F0Track] = CalculateF0Features_MFCC(SpecData, EstF0_InitialTrack, para);    
    otherwise
        error('Wrong feature type!\n');
end

% Find pairwise must-link constraints
para = {};
para.MSL_pd = MSL_pd;                                               % Frequency deviation threshold to decide must links.
AllMustLink = FindAllMustLinks(F0Location, para);                   % Find all must-links that the domain knowledge can provide
PreGroup = FormGroupFromMustLink(AllMustLink);                      % Form groups from must-links
F0Group = MapF0ToGroup(PreGroup, size(F0Location,1));               % Get inverse mapping from F0 to Group

% Calculate pairwise cannot-link constraints
para = {};
para.CNL_opth = CNL_opth*fs/1000/hop;                               % Notelet overlap threshold, if 0 every pair of concurrent pitches have a cannot link
AllCannotLink = FindAllCannotLinks(F0Location, PreGroup, F0Group, para);    % find all cannot-links that the domain knowledge can provide

% Incremental constrained clustering
para = {};
para.DistType = DistType;
[F0Track, CurrMustLink, CurrCannotLink] = IncrementalConstrainedClustering_ConstraintsOnly(F0Feature, F0Track, AllMustLink, AllCannotLink, para);

% Form the final pitch trajectories
EstF0_FinalTrack = FormF0TrajectoryFromF0Track(F0Location, F0Track, length(FrameIndex));% Form the final F0 trajectories
if IfDebug==1
    figure;
    imagesc(1:length(FrameIndex), (1:fftLen/2)*fs/fftLen, 20*log10(abs(SpecData(1:fftLen/2, :))));
    set(gca,'YDir','normal');
    title('Final tracking');
    xlabel('Time (Frame number)');
    ylabel('Frequency (Hz)');
    ylim([0 2000]);
    hold on;
    plot(1:size(EstF0_FinalTrack,2), midi2hz(EstF0_FinalTrack), '.');
end

% Formating notes (non-constant frequencies) of each track
if bNoteForm == 1
    Group = FormGroupFromMustLink(CurrMustLink);                    % form groups from must-links
    GroupTrack = FormGroupTrackFromF0Track(Group, F0Track);         % calculate group track lables
    EstNote = cell(trackNum, 1);
    para = {};
    para.mergeNoteGap = mergeNoteGap*fs/1000/hop;                   % change to frame number
    para.minNoteLength = minNoteLength*fs/1000/hop;                 % change to frame number
    for track = 1:trackNum
        EstNote{track} = FormingNotesForSomeTrack(track, Group, GroupTrack, F0Location, para);
    end

    % Back to F0 estimation (but now it is tracked)
    fprintf('Calculate F0 estimates from notes...\n');
    para = {};
    para.FrameIndex = FrameIndex;
    EstF0_FinalTrack = FormF0AllTrackFromNote(EstNote, para);
    if IfDebug==1
        figure;
        imagesc(1:length(FrameIndex), (1:fftLen/2)*fs/fftLen, 20*log10(abs(SpecData(1:fftLen/2, :))));
        set(gca,'YDir','normal');
        title('Final tracking after note formation');
        xlabel('Time (Frame number)');
        ylabel('Frequency (Hz)');
        ylim([0 2000]);
        hold on;
        plot(1:size(EstF0_FinalTrack,2), midi2hz(EstF0_FinalTrack), '.');
    end        
end

% output multi-pitch tracking result
dlmwrite(outFile, EstF0_FinalTrack', '\t');

fprintf('Done!\n');

